/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pim/mail/MailSyncSource.h"
#include "spds/SyncItemStatus.h"
#include "spds/MailMessage.h"
#include "spds/constants.h"
#include "base/util/utils.h"
#include "base/util/StringBuffer.h"
#include "pim/utils.h"
#include "spds/FolderData.h"
#include "notify/timed_msgbox.h"

#define ASK_REFRESH_MESSAGE TEXT("All the local mails will be deleted before adding the server email. Are you sure to continue?")
#define ASK_REFRESH_TIMEOUT 10

static int call;
static int cnew;
static int cupdated;
static int cdeleted;

static SyncItem *buildSyncItem(MailClientData* m, ArrayList& extras)
{
    MailMessage& msg = m->getEmailData()->getEmailItem();
    /*
    // get the MailExtraData object to retrieve i.e. message-id... It is only the 
    // pointer. don't destroy it!
    const MailExtraData* med = getMailExtraData(extras, msg.getEntryID());
    // extra is empty because the previous file doesn't exist or is empty,
    // or is a slow after previous fast
    if (med) {
        msg.setMessageId(med->getMessageId());
        msg.setHeaders(med->getHeaders());
    }
    else {
        MailExtraData m;
        char *id = createMessageID(msg.getEntryID());

        m.setId(msg.getEntryID());
        m.setMessageId(id);
        extras.add(m);

        msg.setMessageId(id);

        delete [] id;
    }
    */
    wchar_t *id = toWideChar(msg.getEntryID());
    SyncItem *syncItem = new SyncItem(id);
    delete [] id;
    //LOG.debug("MailSyncSource: From: %s", m->getEmailData()->getEmailItem().getFrom());   // TODO: remove this log
    char* toSend = m->getEmailData()->format();
    //LOG.debug(toSend);   // TODO: remove this log
    syncItem->setData(toSend, strlen(toSend) * sizeof(char));              
    syncItem->setSourceParent(m->getParent());
    if (toSend)
        delete [] toSend;
    
    return syncItem;
}

static bool setMailData(SyncItem &item, const wchar_t *parent, MailClientData &m)
{
    int errorCode; 

    char *msg = new char[item.getDataSize() + 1];
    memset(msg, 0, item.getDataSize() + 1);
    memcpy(msg, item.getData(), item.getDataSize());

    errorCode = m.emailData->parse(msg);

    if (errorCode != 0) {
        logError(TEXT("Error parsing message!"));
        return false;
    }
    m.setParent(parent);

    delete [] msg;
    return true;
}

static bool setFolderData(SyncItem &item, /*const wchar_t *parent*/ FolderData &folder)
{
    int errorCode; 

    char *msg = new char[item.getDataSize() + 1];
    memset(msg, 0, item.getDataSize() + 1);
    memcpy(msg, item.getData(), item.getDataSize());

    errorCode = folder.parse(msg);

    if (errorCode != 0) {
        logError(TEXT("Error parsing message!"));
        return false;
    }
    
    delete [] msg;
    return true;
}


// set all extra data
//
static inline void setExtraData(ArrayList& extras,
                                const wchar_t *id,
                                const char *msgid, 
                                const char * extraHeaders)
{
    MailExtraData med;

    char *entryID = toMultibyte(id);
    med.setId(entryID);
    med.setMessageId(msgid);
    med.setHeaders(extraHeaders);
    extras.add(med);
    delete [] entryID;
}


MailSyncSource::MailSyncSource(const wchar_t* name) : SyncSource(name)
{    
    path = wstrdup(TEXT("\\"));
    s_folderToSync = NULL;   
    syncItem = NULL;
    isMailInOutbox = false;
    isMailInInbox = false;
    mailClient2Server = NULL;
    mailServer2Client = NULL;
    failedSendMailInOutbox = false;
}

MailSyncSource::MailSyncSource(MailSyncSource& s) : SyncSource(s)
{
    setLastSync(s.getLastSync());
    setNextSync(s.getNextSync());
    setRemoteURI(s.getRemoteURI());
    setType(s.getType());
    setLastAnchor(s.getLastAnchor());
    setNextAnchor(s.getNextAnchor());
    setErrorHandler(s.getErrorHandler());
    setFilter(s.getFilter());

    setPath(s.path);
    setFolderToSync(s.s_folderToSync);

    // These are not copied from the other instance
    isMailInOutbox = false;
    isMailInInbox = false;
    failedSendMailInOutbox = false;
    mailServer2Client=NULL;
    mailClient2Server=NULL;
    extras.clear();
}

MailSyncSource::~MailSyncSource() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;
    }
    if (s_folderToSync) {
        delete [] s_folderToSync; s_folderToSync = NULL;
    }
    if (path) {
        delete [] path; path = NULL;
    }
    if (mailServer2Client)
        delete mailServer2Client;
    if (mailClient2Server)
        delete mailClient2Server;
}

void MailSyncSource::setPath(const wchar_t* p) {
    if (path)
        delete [] path;

    if (p == NULL)
        path = wstrdup(TEXT(""));
    else
        path = wstrdup(p);
}

const wchar_t* MailSyncSource::getPath() {
    return path;
}

void MailSyncSource::setFolderToSync(const wchar_t* t){
    if(s_folderToSync)
        delete [] s_folderToSync;
    s_folderToSync = wstrdup(t);
}

const wchar_t* MailSyncSource::getFolderToSync() {
    return s_folderToSync;
}

bool MailSyncSource::isInFolderToSync(const wchar_t* toCheck) {
    return wcsstr(s_folderToSync, toCheck) != NULL ? true : false;
}

void MailSyncSource::setIsMailInOutbox(bool value){
    isMailInOutbox = value;
    
}

bool MailSyncSource::getIsMailInOutbox() {
    return isMailInOutbox;
}

void MailSyncSource::setIsMailInInbox(bool value){
    isMailInInbox = value;    
}

bool MailSyncSource::getIsMailInInbox() {
    return isMailInInbox;
}

void MailSyncSource::setFailedSendMailInOutbox(bool value){
    failedSendMailInOutbox = value;    
}

bool MailSyncSource::getFailedSendMailInOutbox() {
    return failedSendMailInOutbox;
}

/*
 * Return the first SyncItem of all. It is used in case of slow 
 * sync and retrieve the entire data source content.
 */
SyncItem* MailSyncSource::getFirstItem()
{
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    //
    // initialize the flag
    //
    isMailInInbox = false;

    if ( mailClient2Server->setSync(SYNC_SLOW, getPath(), extras) != S_OK ){
        logError(TEXT("getFirstItem: error in MailClient2Server::setSync"));
        return NULL;
    }
    
    MailClientData* m = mailClient2Server->getFirstItem();
    
    if (m) {              
        logInfo(TEXT("Subject first all mail: -%S-"), m->getEmailData()->getEmailItem().getSubject());   
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_NONE);
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No mail get"));
    }
    
    return syncItem;    
}

/*
* Return the next SyncItem of all. It is used in case of slow or refresh sync 
* and retrieve the entire data source content.
*/

SyncItem* MailSyncSource::getNextItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextItem();   
    if (m) {      
        logInfo(TEXT("Subject other all mail: -%S-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);                        
        syncItem->setState(SYNC_STATE_NONE);
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
                
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No mail get"));
    }

    return syncItem;  

}

SyncItem* MailSyncSource::getFirstItemKey() {
        
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    } 
    //
    // initialize the flag
    //
    setIsMailInOutbox(false);
    
    mailClient2Server->setFolderToSync(s_folderToSync);

    if ( mailClient2Server->setSync(SYNC_SLOW, getPath(), extras) != S_OK ){
        logError(TEXT("getFirstItemKey: error in MailClient2Server::setSync"));
        return NULL;
    }

    MailClientData* m = mailClient2Server->getFirstItem();
    
    if (m) {
        mailServer2Client->setFolderToSync(s_folderToSync);
        mailServer2Client->deleteMessage(m);        
        delete m; m = NULL;

        while((m = mailClient2Server->getNextItem()) != NULL) {
            mailServer2Client->deleteMessage(m);
            delete m; m = NULL;
        }
    }

    return NULL;
    /*
    if (m) {
        wchar_t *id = toWideChar(m->getEmailData()->getEmailItem().getEntryID());
        syncItem = new SyncItem(id);
        delete [] id;
        delete m; m = NULL;
    }
    
    return syncItem;    
    */
    
}

SyncItem* MailSyncSource::getNextItemKey() {
    
    /*
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextItem();   
    if (m) {
        wchar_t *id = toWideChar(m->getEmailData()->getEmailItem().getEntryID());
        syncItem = new SyncItem(id);
        delete [] id;
        delete m; m = NULL;
    }
    
    return syncItem;  
    */
    return NULL;
}

/*
* This method is always called by the sync manager. So the setModifiedItemsPOOM is always called
*/

SyncItem* MailSyncSource::getFirstNewItem() {

    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    //
    // initialize the flag
    //
    setIsMailInOutbox(false);
    
    mailClient2Server->setFolderToSync(s_folderToSync);

    if ( mailClient2Server->setSync(SYNC_TWO_WAY, getPath(), extras) != S_OK ){
        logError(TEXT("getFirstNewItem: error in MailClient2Server::setSync"));
        return NULL;
    }
    
    MailClientData* m = mailClient2Server->getFirstNewItem();   
    
    if (m) {      
        logInfo(TEXT("Subject first new mail: -%S-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_NEW);
        //syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No new mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextNewItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextNewItem();   
    
    if (m) {      
        logInfo(TEXT("Subject other new mail: -%S-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_NEW);
        //syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No other new mail"));
    }
    /*
    if (m) {      
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            char *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }
        wchar_t *id = toWideChar(emailItem.getEntryID());
        syncItem = new SyncItem(id);
        delete [] id;
        
        LOG.debug("MailSyncSource: From: %ls\n", m->getEmailData()->getEmailItem().getFrom());        
        char* toSend = m->getEmailData()->format();
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));
        delete [] toSend;

        syncItem->setState(SYNC_STATE_NEW);
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No other new mail"));
    }
    */
    return syncItem;  
}

SyncItem* MailSyncSource::getFirstUpdatedItem() {
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
        
    MailClientData* m = mailClient2Server->getFirstUpdatedItem();       

    if (m) {      
        logInfo(TEXT("Subject first updated mail: -%S-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_UPDATED);
        //syncItem->setSourceParent(m->getParent());
       
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No updated mail"));
    }
  /*  
    if (m) {      
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            char *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }
        wchar_t *id = toWideChar(m->getEmailData()->getEmailItem().getEntryID());
        syncItem = new SyncItem(id);
        
        char* toSend = m->getEmailData()->format();
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));
        delete [] toSend;
        syncItem->setState(SYNC_STATE_UPDATED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        
        logInfo(TEXT("Subject first updated mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No updated mail"));
    }
*/
    return syncItem;  
}

SyncItem* MailSyncSource::getNextUpdatedItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getNextUpdatedItem();   
    
    if (m) {      
        logInfo(TEXT("Subject other updated mail: -%S-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_UPDATED);
        //syncItem->setSourceParent(m->getParent());
        
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No other updated mail"));
    }
    /*
    if (m) {      
        
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            char *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }

        wchar_t *id = toWideChar(emailItem.getEntryID());
        syncItem = new SyncItem(id);
        delete [] id;
        

        char* toSend = m->getEmailData()->format();
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));
        delete [] toSend;
        syncItem->setState(SYNC_STATE_UPDATED);
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No other updated mail"));
    }
    */
    return syncItem;  
}

SyncItem* MailSyncSource::getFirstDeletedItem() {
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getFirstDeletedItem();   
    
    if (m) {
        wchar_t *id = toWideChar(m->getEmailData()->getEmailItem().getEntryID());
        syncItem = new SyncItem(id);
        syncItem->setState(SYNC_STATE_DELETED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        
        logInfo(TEXT("ID of deleted mail: -%s-"), id);
        delete m; m = NULL;
        delete [] id;
    
    } else {
        logInfo(TEXT("No deleted mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextDeletedItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getNextDeletedItem();   
    
    if (m) {
        wchar_t *id = toWideChar(m->getEmailData()->getEmailItem().getEntryID());
        syncItem = new SyncItem(id);     
        syncItem->setState(SYNC_STATE_DELETED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        logInfo(TEXT("ID of other deleted mail: -%s-"), id);

        delete m; m = NULL;
        delete [] id;

    } else {
        logInfo(TEXT("No other deleted mail"));
    }

    return syncItem;  
}

void MailSyncSource::setItemStatus(const wchar_t* key, int status) {
    
    if (wcsstr(key, TEXT("O/")) != NULL) {
        if (status == 500 || status == 418 || status == 419) {
            failedSendMailInOutbox = true;            
        }
    }

    logDebug(TEXT("key: %s, status: %i"), key, status);
}

int MailSyncSource::addItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    MailClientData m;
    FolderData folder;

    wchar_t* id = NULL;
    wchar_t tt[3];
    
    mailServer2Client->setFolderToSync(s_folderToSync);
    
    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());
    }
    
    char* itemData = (char*)item.getData();
    if(!itemData){
        ret = 500;
        goto finally;
    }
    char* posFolder = strstr(itemData, "Folder>");
    if ((posFolder != NULL) && ((posFolder - itemData) < 10))
        if (!setFolderData(item, folder))
        {
            ret = 500;
            goto finally;
        }
        else
        {
            ret = 200;
            if (wcsicmp(folder.getRole(), TEXT("Inbox")) == 0)
                item.setKey(TEXT("I"));
            else if (wcsicmp(folder.getRole(), TEXT("Outbox")) == 0)
                item.setKey(TEXT("O"));
            else if (wcsicmp(folder.getRole(), TEXT("Drafts")) == 0)
                item.setKey(TEXT("D"));
            else if (wcsicmp(folder.getRole(), TEXT("Sent")) == 0)
                item.setKey(TEXT("S"));
            else
                ret = 500;            
            goto finally;
        }
    else{
        if (!isInFolderToSync(tt))
            goto finally;
        if (!setMailData(item, tt, m))
            goto finally;
    }

    
    id = mailServer2Client->addMessage(&m);
    if (!id) {
        logError( TEXT("Error in MailServer2Client::addMessage") );
        goto finally;
    }
    else if (wcscmp(id, ITEM_ARLEADY_EXISTS) == 0) {
        ret = STC_ALREADY_EXISTS; 
        logDebug( TEXT("Mail arleady exists"));
        goto finally;
    }

    // Ok: set the new key and the status code
    ret = STC_ITEM_ADDED;        
    item.setKey(id);
    MailMessage& msg = m.getEmailData()->getEmailItem();
    logInfo(TEXT("Mail added: -%S-"), msg.getSubject());
    
    //char * hdrs = msg.getHeaders();
    //setExtraData(extras, id, msg.getMessageId(), hdrs);
    //delete [] hdrs;

    // Set new mail in Inbox flag
    if (isMailInInbox == false && wcscmp(tt, TEXT("I")) == 0) 
        isMailInInbox = true;

finally:
    
    return ret;
}

int MailSyncSource::updateItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    wchar_t tt[3];
    MailClientData m; 
    int errorCode = 0;
    HRESULT hr = S_OK;
    char* tmp = NULL;
    wchar_t* toUse = NULL;

    mailServer2Client->setFolderToSync(s_folderToSync);
          
    if (!isInFolderToSync(tt))
        goto finally;     
    
    if (!setMailData(item, tt, m))
        goto finally;
           
    hr = mailServer2Client->updateMessage(&m);
    
    if (hr == S_OK) {        
        ret = STC_OK;     
        logInfo(TEXT("Mail updated: -%-"), m.getEmailData()->getEmailItem().getSubject());
    }    
    
finally:
    
    return ret;
}

int MailSyncSource::deleteItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    wchar_t tt[3];

    mailServer2Client->setFolderToSync(s_folderToSync);
    MailClientData m;

    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());    
    }

    m.setParent(tt);
    char *id = toMultibyte(item.getKey());
    m.emailData->getEmailItem().setEntryID(id);
    delete [] id;
    
    HRESULT hr = mailServer2Client->deleteMessage(&m);
    
    if (hr == S_OK) {        
        ret = STC_OK;          
        logInfo(TEXT("Mail deleted: -%-"), item.getKey());
    }
        
    return ret;
}

int MailSyncSource::beginSync() {
    // Init mailServer2Client
    if (mailServer2Client)
        delete mailServer2Client;

    mailServer2Client = new MailServer2Client();
    
    if(!mailServer2Client)
        return 500;

    // Init mailClient2server
    if (mailClient2Server)
        delete mailClient2Server;

    mailClient2Server = new MailClient2Server();
    mailClient2Server->setFolderToSync(s_folderToSync);
        
    if(!mailClient2Server)
        return 500;
    
    if (getSyncMode() == SYNC_REFRESH_FROM_SERVER) {
        int val = TimedMessageBox(NULL, ASK_REFRESH_MESSAGE, TEXT("Alert"),
                             MB_YESNO | MB_ICONQUESTION, ASK_REFRESH_TIMEOUT * 1000); // default ASK_TIMEOUT = 10

        if (val == IDYES) {
            LOG.debug("User said yes to delete local mail and refresh.");            
        }
        else {
            LOG.debug("User said no to delete local mail and refresh");     
            return -1;
        }
    }
    return 0;
}

int MailSyncSource::endSync() {
    
    // Get a new mailClient2server
    if (mailClient2Server)
        delete mailClient2Server;

    mailClient2Server = new MailClient2Server();
    mailClient2Server->setFolderToSync(s_folderToSync);
    mailClient2Server->setSync(ONLY_ALL_MESSAGE_LIST, getPath(), extras);    
    mailClient2Server->writeCurrentMails(getPath(), extras);

    return 0;
}

MailSyncSource& MailSyncSource::operator=(MailSyncSource& s) {
    if (this != &s) {
        setSyncMode(s.getSyncMode());
        setLastSync(s.getLastSync());
        setNextSync(s.getNextSync());
        setRemoteURI(s.getRemoteURI());
        setType(s.getType());
        setLastAnchor(s.getLastAnchor());
        setNextAnchor(s.getNextAnchor());
        setErrorHandler(s.getErrorHandler());

        setFilter(s.getFilter());
        setPath(s.getPath());
        setFolderToSync(s.getFolderToSync());

        // This attributes are not copied from the other source
        setIsMailInOutbox(false);
        setIsMailInInbox(false);
        mailClient2Server = NULL;
        mailServer2Client = NULL;
        extras.clear();
    }
    return *this;
}

ArrayElement* MailSyncSource::clone() {
    MailSyncSource* s = new MailSyncSource(getName());
    *s = *this;
    return s;
}



